
from .version import version

__version__ = version
__license__ = 'BSD3'
__author__ = 'Kyle B. Westfall'
__maintainer__ = 'Kyle B. Westfall'
__email__ = 'westfall@ucolick.org'
__copyright__ = '(c) 2014-2022, SDSS-IV/MaNGA Pipeline Group'
__credits__ = ['Kyle B. Westfall', 'Brett H. Andrews', 'Jorge Barrera-Ballesteros',
               'Francesco Belfiore', 'Matthew A. Bershady', 'Kevin Bundy', 'Joel R. Brownstein',
               'Michele Cappellari', 'Brian Cherinka', 'Lodovico Coccato', 'Niv Drory',
               'Cheng Du', 'Eric Emsellem', 'Daniel Goddard', 'Xihan Ji', 'David R. Law',
               'Niu Li', 'Claudia Maraston', 'Karen Masters', 'Héctor Javier Ibarra Medel',
               'Taniya Parikh', 'Sebastián F. Sánchez', 'José R. Sánchez-Gallego',
               'Adam Schaefer', 'Shravan Shetty', 'Daniel Thomas', 'Christy A. Tremonti',
               'Anne-Marie Weijmans', 'Renbin Yan', 'Meng Yang', 'Zheng Zheng', 'Shuang Zhou']

#def short_warning(message, category, filename, lineno, file=None, line=None):
#    """
#    Return the format for a short warning message.
#    """
#    return ' %s: %s\n' % (category.__name__, message)
#
#import warnings
#warnings.formatwarning = short_warning

